#include <math.h>

#include "StdAfx.h"
#include "application.h"
#include "Universe.h"
#include "tick.h"
#include "TextureManager.h"

#include <gmtl/QuatOps.h>
#include <gmtl/Quat.h>
#include <gmtl/Vec.h>
#include <gmtl/Matrix.h>
#include <gmtl/Generate.h>
#include <gmtl/Output.h>

#include <vrj/Kernel/Kernel.h>

#include <GL/gl.h>

bool RENDER_REFLECT = true; //reflections on - B
bool RENDER_SHADOWS = true; //shadows on - X
bool drawShadowVolumes = false; //draw the shadow volumes - C
bool drawThings = true; //draw the things in the scene - V
bool linkUser = false; //connect the camera and avatar (ie first person perspective)
bool drawFakeShadows = true; //draw fake (projected) shadows


Application::Application(Universe *uni)
{
	universe = uni;

	// Controller is the "user" representation
	controller = uni->getController();
	countdown = 0;
}

Application::~Application(void)
{
	delete universe;
}

void Application::init()
{
	std::vector<vrj::User*> users = mKernel->getUsers();
	mUser = users[0];

	mWand.init("VJWand");
	mHead.init("VJHead");
	keyboard.init("VJKeyboard");
    mIncVelocityButton.init("VJButton0");     // Button for velocity
    mDecVelocityButton.init("VJButton1");
    mStopButton.init("VJButton2");            // Button to stop
    mJoystick[0].init("VJAnalog0");           // 2D Joystick
    mJoystick[1].init("VJAnalog1");           // 2D Joystick

	gmtl::setScale(nav_mat, gmtl::Vec3f(0.3f, 0.3f, 0.3f)); // This scale sets up CAVE->world scaling

    // Defaults for joystick navigation
    mSpeedMax=2.0;
    mSpeedExponent=2.0;
    mRotateSpeedMax=5.0;
    mRotateSpeedExponent=2.0;
}

void Application::apiInit()
{
	universe->init();

	m_lastTick = 0;
	m_bPaused = false;
	m_lastTick = Tick::getTick();
}

void Application::contextInit()
{
	TextureManager::getInstance()->loadTexturesToGL();
	//initialise OpenGL
	//Use depth buffering for hidden surface elimination.
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glEnable(GL_STENCIL);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LIGHTING);

	//we are going to use texture mapping so enable 
	glEnable(GL_TEXTURE_2D);
}

void Application::bufferPreDraw()
{
}

void Application::preFrame()
{
	//SaneCubes rip off
	gmtl::Matrix44f    rot_mat;

	// -----------------------------------------------------------------------
	// joystick-based navigation...
	float x, y;
	x = mJoystick[0]->getData();
	y = mJoystick[1]->getData();

	//This puts in a small "deadzone" in the centre of the joystick, but still allows slow movement 
	if (x<0.45)  x += 0.05;
	else if (x<0.55)x = 0.5;
	else x -= 0.05;
	if (y<0.45) y += 0.05;
	else if (y<0.55) y = 0.5;
	else y -= 0.05;

	gmtl::Matrix44f head_tracker_mat= mUser->getHeadPosProxy()->getData();

	// transform in cave local coordinates, 
	// i.e. rotate about centre of reactor, and translate in direction
	// of centre screen

	float rotateSpeed= mRotateSpeedMax*(x-0.5);
	if (mRotateSpeedExponent!=1.0) {
		if (rotateSpeed>=0)
			rotateSpeed = pow(rotateSpeed, mRotateSpeedExponent);
		else
			rotateSpeed = -pow(-rotateSpeed, mRotateSpeedExponent);
	}

	gmtl::EulerAngleXYZf euler( 0, gmtl::Math::deg2Rad(-rotateSpeed), 0 );

	gmtl::setRot(rot_mat, euler);


	//Head position in world coordinates
	gmtl::Matrix44f head_world_mat;
	head_world_mat = nav_mat*head_tracker_mat;

	// Calculate a rotation about the head
	gmtl::Matrix44f pre_trans_mat;
	gmtl::setTrans(pre_trans_mat, gmtl::Vec3f(-head_world_mat(0,3),
		-head_world_mat(1,3), 
		-head_world_mat(2,3)));

	gmtl::Matrix44f post_trans_mat;
	gmtl::setTrans(post_trans_mat, gmtl::Vec3f(head_world_mat(0,3),
		head_world_mat(1,3), 
		head_world_mat(2,3)));

	float speed = -((y-0.5)/2.0)*mSpeedMax;
	if (mSpeedExponent!=1.0) {
		if (speed>=0)
			speed = pow(speed, mSpeedExponent);
		else
			speed = -pow(-speed, mSpeedExponent);
	}

	// Calculate a translation along the ground
	gmtl::Matrix44f nav_trans_mat;
	gmtl::setTrans(nav_trans_mat, gmtl::Vec3f(-head_world_mat(2,0)*speed, 0, head_world_mat(2,2)*speed));

	// Calculate the new navigation matrix
	nav_mat = nav_trans_mat*post_trans_mat*rot_mat*pre_trans_mat*nav_mat;

	// Move the avatar to our position - get new head position in world coordinates
	head_world_mat = nav_mat*head_tracker_mat;

	controller->setX(head_world_mat(0,3));
	controller->setZ(head_world_mat(2,3));

	//keyboard input - turn on/off effects
	if (countdown == 0)
	{
#if 0
		if(keyboard->keyPressed(gadget::KEY_W))
		{
			controller->onKeyPress('w', 0, 0);
			//countdown = 10;
		}
		if(keyboard->keyPressed(gadget::KEY_S))
		{
			controller->onKeyPress('s', 0, 0);
			//countdown = 10;
		}
		if(keyboard->keyPressed(gadget::KEY_A))
		{
			controller->onKeyPress('a', 0, 0);
			//countdown = 10;
		}
		if(keyboard->keyPressed(gadget::KEY_D))
		{
			controller->onKeyPress('d', 0, 0);
			//countdown = 10;
		}
#endif
		if(keyboard->keyPressed(gadget::KEY_X))
		{
			RENDER_SHADOWS = !RENDER_SHADOWS;
			countdown = 10;
		}
		if(keyboard->keyPressed(gadget::KEY_C))
		{
			drawShadowVolumes = !drawShadowVolumes;
			countdown = 10;
		}
		if(keyboard->keyPressed(gadget::KEY_V))
		{
			drawThings = !drawThings;
			countdown = 10;
		}
		if(keyboard->keyPressed(gadget::KEY_B))
		{
			RENDER_REFLECT = !RENDER_REFLECT;
			countdown = 10;
		}
		if(keyboard->keyPressed(gadget::KEY_N))
		{
			linkUser = !linkUser;
			countdown = 10;
		}
	}
	else
		--countdown;

	// get the current tick value
	unsigned int tick;
	tick = Tick::getTick();

	// calculate the amount of elapsed seconds
	float elapsedSeconds;
	elapsedSeconds = (float)(tick - m_lastTick) / 1000.0f;

	// update the model if not paused
	if(!m_bPaused)
	{
		universe->update(elapsedSeconds);
	}

	// current tick will be last tick next round
	m_lastTick = tick;
}

void Application::intraFrame()
{
}

void Application::postFrame()
{
}

void Application::draw()
{
	//frameStart();
 	glStencilMask(~0);
	//clear the screen
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	glLoadIdentity();

   glPushMatrix();

   // --- Push on Navigation matrix for the user --- //
   glMultMatrixf(gmtl::makeInvert(nav_mat).mData);
   glPushMatrix();
//	glTranslatef(4, 6.6, 4);
	universe->render();
	glPopMatrix();
	glPopMatrix();

}
