#include <stdlib.h>
#include <time.h>
#include <iostream>
#include "StdAfx.h"
#include "universe.h"
#include "SimpleMap.h"
#include "Camera.h"
#include "Light.h"
#ifdef VC71
#include <GL/glut.h>
#else
#include "glut.h"
#endif

#include "Box1.h"
#include "Vertex.h"
#include "mirror.h"
#ifdef VC71
#else
#endif
#include "FurnitureThing.h"

extern bool RENDER_REFLECT;
extern bool RENDER_SHADOWS;
extern bool drawShadowVolumes;
extern bool drawThings;
extern bool linkUser;
extern bool drawFakeShadows;

Universe::Universe(void)
{
	//create new user and light
	user = new User(this);
	GLfloat light_position0[] = {-3.0, 2.0, 8.0, 0.0};
	light = new Light(GL_LIGHT0, light_position0);
}

void Universe::addRandomBoxes(int n)
{
	//adds n randomly positioned boxes to the scene
	//note this is a hack since the size of the map is hardcoded in
	srand((unsigned)time(NULL));

	for (int i = 0; i < n; ++i)
	{
		float x = ((rand() % 200)/10.0) - 5;
		float y = (rand() % 20)/10.0;
		float z = ((rand() % 200)/10.0) - 5;
		Thing *box = new FurnitureThing(x, y, z, Box1::getInstance());
		addThing(box);
	}
}

Universe::~Universe(void)
{
	for (unsigned int i = 0; i < things.size(); ++i)
	{
		delete things[i];
	}
	delete map;
	delete user;
}

void Universe::render()
{
	user->getCamera()->render();

	//render any reflectors
	for (unsigned int i = 0; i < things.size(); ++i)
	{
		if (things[i]->reflects())
			things[i]->render();
	}

	//render the non-reflecting scene
	renderNonReflect(false, false);
}

void Universe::update(float elapsedSeconds)
{
	for (unsigned int i = 0; i < things.size(); ++i)
	{
		things[i]->update(elapsedSeconds);
	}

	user->update(elapsedSeconds);
}


void Universe::renderNonReflect(bool showAvatar, bool reflected)
{
	//**first pass (ambient lighting)**

	//turn on ambient lighting
	float ambient[] = {0.2, 0.2, 0.2, 1.0};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
	light->enable(false);
	renderScene(showAvatar);

	if(RENDER_SHADOWS)
	{
		//**render shadow volumes**
		//enable the stencil buffer
		glEnable(GL_STENCIL_TEST);
		if (reflected)
		{
			glStencilFunc(GL_EQUAL, 0x80, 0x80);
		}
		else
		{
			glStencilFunc(GL_ALWAYS, 0, ~0);
		}
		glStencilMask(0x7f);
		//disable writes to the colour and depth buffers
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		glDepthMask(GL_FALSE);

		//render the backs of the shadow volumes, into the stencil buffer by incrementing the
		//count there
		if (!reflected) //we are in this state already
			glCullFace(GL_FRONT);
		else
			glCullFace(GL_BACK);
		glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
		renderShadowVolumes();

		//render the fronts of the shadow volumes, into the stencil buffer by decrementing the
		//count there
		if (!reflected)
			glCullFace(GL_BACK);
		else
			glCullFace(GL_FRONT);
		glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
		renderShadowVolumes();

		//**second pass (diffuse lighting)**
		//enable writing to the colour buffer
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

		//draw only if the value in the stencil buffer is 0
		if (reflected)
			glStencilFunc(GL_EQUAL, 0x80, ~0);
		else
			glStencilFunc(GL_EQUAL, 0, 0x7f);

		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	}

	//enable depth testing, passes if equal to what is there already
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_EQUAL);

	//blend this pass with the last one
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);

	//turn off ambient lighting
	float ambientOff[] = {0, 0, 0, 0};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientOff);

	light->enable(true);
	renderScene(showAvatar);

	//normal depth testing
	glDepthFunc(GL_LEQUAL);

	//projected shadow setups
	glEnable(GL_STENCIL_TEST);
	if (reflected)
		glStencilFunc(GL_EQUAL, 0x80, ~0);
	else
		glStencilFunc(GL_EQUAL, 0, 0x7f);
	glStencilMask(0x7f);
	glStencilOp(GL_KEEP, GL_INCR, GL_INCR);

	renderFakeShadows();

	//disable the stencil buffer
	glDisable(GL_STENCIL_TEST);
	//disable blending
	glDisable(GL_BLEND);

}

void Universe::renderFakeShadows()
{
	//set up GL bits
	glPolygonOffset(-1.0, 1.0);
	glEnable(GL_POLYGON_OFFSET_FILL);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glColor4f(0.0, 0.0, 0.0, 0.5);

	//render the objects
	if (drawFakeShadows)
	{
		for (unsigned int i = 0; i < things.size(); ++i)
		{
			things[i]->renderProjectedShadows(light->getPosition());
		}
	}

	//render the user
	if (drawFakeShadows || RENDER_SHADOWS)
		user->renderProjectedShadows(light->getPosition());

	//tear down GL bits
	glDisable(GL_POLYGON_OFFSET_FILL);
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
}

void Universe::renderScene(bool showAvatar)
{
	light->render();
	map->render();
	for (unsigned int i = 0; i < things.size(); ++i)
	{
		if (drawThings)
		{
			if (!things[i]->reflects())
				things[i]->render();
		}

		//for debugging
		if (drawShadowVolumes)
		{
			glCullFace(GL_FRONT);
			GLfloat colour[] = {1.0, 0.0, 1.0, 1.0};
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colour);
			things[i]->renderShadowVolumes(light->getPosition());

			glCullFace(GL_BACK);
			GLfloat colour2[] = {1.0, 1.0, 0.0, 1.0};
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colour2);
			things[i]->renderShadowVolumes(light->getPosition());
		}
	}

	//for debugging
	if (drawShadowVolumes)
	{
		glCullFace(GL_FRONT);
		GLfloat colour[] = {1.0, 0.0, 1.0, 1.0};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colour);
		user->renderShadowVolumes(light->getPosition());

		glCullFace(GL_BACK);
		GLfloat colour2[] = {1.0, 1.0, 0.0, 1.0};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colour2);
		user->renderShadowVolumes(light->getPosition());
	}

	if (showAvatar || !linkUser)
	{
		user->render();
	}
}

void Universe::renderShadowVolumes()
{
	for (unsigned int i = 0; i < things.size(); ++i)
	{
		things[i]->renderShadowVolumes(light->getPosition());
	}
	user->renderShadowVolumes(light->getPosition());
}

void Universe::init()
{
	user->init();
	addRandomBoxes(20);
	Thing *mirror = new Mirror(&Vertex(0, 0, 0), &Vertex(4, 1.5, 4), this);
	addThing(mirror);
	map = new SimpleMap(-5, 15, -5, 15);
	light->init();
	light->enable(true);
	for (unsigned int i = 0; i < things.size(); ++i)
	{
		things[i]->init();
	}
}
